; tree-view.lisp

; =========================================================================================
; based on
; http://osdir.com/ml/mcclim-devel/2009-08/msg00010.html
; http://www.cs.cmu.edu/afs/cs/project/ai-repository/ai/lang/lisp/gui/clim/clim_2/browsers/
;
;;; Written by Jeff Morrill (jmorrill@bbn.com), December 92.
;;; It is provided "as is" without express or implied warranty.
;;; Thanks to Scott McKay for solving an incremental redisplay problem.
; =========================================================================================

(in-package cw-treeview)

;************************************************************
; 1) ICONS AND GRID
;************************************************************
(defvar icon 'plus)     ; 'triangle 'triangle2
(defvar grid t)         ; nil
(defvar dashed-lines t) ; nil

(define-presentation-type icon ())
(define-presentation-method highlight-presentation ((type icon) r s st) :unhighlight)
(define-presentation-method present (o (type icon) s v &key) (disp-tree o ptype s (indentation o)))

(defmethod toggle (n) (setf (show-children n) (not (show-children n))))

; when doing all calculationsts with d, there is no need for scaling
(define-symbol-macro d (STREAM-CHARACTER-WIDTH s #\m))

; GRID
(defun spc (s) (stream-increment-cursor-position s (* 1.5 d) nil))
(defun bar% (s) (draw-line* s 0 0 0 d :line-dashes dashed-lines :ink +gray+))  
(defun bar (s) 
  (multiple-value-bind (x y) (stream-cursor-position s) (with-translation (s (+ x (* .5 d))  y) (bar% s)))
  (stream-increment-cursor-position s (* 1.5 d) nil))

(defmethod grid (s i n)
  (let* ((spaces (indent n))
         (bars (- i spaces))) 
    ;1) do initial spaces
    (dotimes (x spaces) (spc s))
    ;2) do bars
    (dotimes (x bars) (if grid (bar s) (spc s)))))

; POINTER
(defun pointr% (s) (draw-lines* s (list 0 (* .3 d) 0 d 0 d (* .7 d) d) :line-dashes dashed-lines :ink +gray+))
(defun pointr (s)
  (multiple-value-bind (x y) (stream-cursor-position s)
    (with-translation (s (+ x (* .5 d)) (+ y (* -.5 d)))
;      (pointr% s))))
      (when grid (pointr% s))))) ; to suppress pointer

; ICONS
;;;plus - remove visible background pointer
(defun rect (s) (draw-rectangle* s 0 0 d d :filled nil))
(defun m (s) (draw-line* s (* d .2) (* d .5) (* d .8) (* d .5)))
(defun p (s) (m s) (draw-line* s (* d .5) (* d .2) (* d .5) (* d .8)))
(defun plus (s x y o) (with-translation (s x y) (if o (m s) (p s)) (rect s)))

;;;triangles
(defun tri-m (s) (draw-polygon* s (list 0 0 d 0 (* .5 d) d) :ink +black+))
(defun tri-p (s) (draw-polygon* s (list 0 0 d (* .5 d) 0 d)))   ; with rotation !!
(defun triangle  (s x y o) (with-translation (s x y)          (if o (tri-m s) (tri-p s))))
(defun triangle2 (s x y o) (with-translation (s x y) (rect s) (if o (tri-m s) (tri-p s))))

(defun draw-icon (s group)
  (with-output-as-presentation (s group 'icon)
    (let ((open-p (show-children group)))
      (multiple-value-bind (x y) (stream-cursor-position s)
        (funcall icon s x y open-p)))))

;************************************************************
; 2) CREATE CHILDREN INSTANCES
;************************************************************
(defclass item ()
  ((name :accessor name :initarg :name)  ; ev name
   (indent :accessor indent :initarg :indent :initform 0)))

(defclass leaf (item) ())  ; leaf shall/may not be a superclass of node 

(defclass node (item) 
  ((children :accessor children :documentation "list of children with attributes such as indentation, node or leaf, youngest-child")
   (youngest-child :accessor youngest-child :initarg :youngest-child :initform t)
   (show-children :initform nil :initarg :show-children :accessor show-children :documentation "boolean")))  ; change default to t <---   rename show-children <----

(defmethod item-name (n) n)
(defgeneric node-p (n) (:documentation "test if child is node or leaf"))
(defgeneric get-children (n) (:documentation "get a list of children"))

(defmacro define-node-methods (&key  ;short keywords
                     ((:nc node-class) 'node) 
                     ((:cc children-class) 'string) 
                     ((:nn item-name-form) '(name n)) 
                     ((:ln leaf-name-form) '(name n)) 
                     ((:gc get-children-form) '(gethash (name n) nodes)) 
                     ((:cp node-p-form) '(gethash n nodes)) 
                     ((:cy childnode-is-youngestsibling-form) 'string=))
  "Define subclasses of node and leaf CLASSES with corresponding METHODS to get the childs of a node.
   All these classes and methods can also be defined, subclassed, etc. in the normal way.
   Keyword arguments:
   :nc node class
   :cc children class
   :nn node name
   :ln leaf name
   :gc  supply a form that returns a list of a node's children    ### ev cl  for children-list <----
   :cp supply a form to test if a child is node or leaf
   :cy supply a function to test for equality o child nodes
   See view directory for an example usage"
   (let ((leaf-class (intern (#~s'NODE'LEAF'i (symbol-name node-class)) *package*)))
     `(progn
        ;classes
        (unless (eql ',node-class 'node)
          (defclass ,node-class (node) ())
          (defclass ,leaf-class (leaf) ()))
        ;methods
        (defmethod item-name ((n ,node-class)) ,item-name-form)
        (defmethod item-name ((n ,leaf-class)) ,leaf-name-form)
        (defmethod get-children ((n ,node-class)) ,get-children-form)
        (defmethod node-p ((n ,children-class)) ,node-p-form) ; the child is a node if it has children
        (defmethod childnode-is-youngestsibling ((n ,children-class) ch) (and (node-p n) (,childnode-is-youngestsibling-form n (car (last ch)))))
        (defmethod children :before ((n ,node-class))
          "create children-instances"
          (unless (slot-boundp n 'children) 
            (let ((children (get-children n))
                  (sp (indent n)))
              (setf (children n) (mapcar 
                              (lambda (x) 
                                (if (node-p x) 
                                  (if (youngest-child n) ; if this (parent)-node is youngest sibling
                                    (if (childnode-is-youngestsibling x children)
                                      (make-instance ',node-class :name x :indent (1+ sp) :youngest-child t)
                                      (make-instance ',node-class :name x :indent (1+ sp) :youngest-child nil))
                                    (make-instance ',node-class :name x :indent sp :youngest-child nil))
                                  (if (youngest-child n)
                                    (make-instance ',leaf-class :name x :indent (1+ sp))
                                    (make-instance ',leaf-class :name x :indent sp))))
                              children))))))))

;************************************************************
; 3) DISPLAY ITEMS
;************************************************************
;nodes
(defmethod disp-tree :around ((item node) pt s indent)
  "This displays grid, pointer, icon and children of a node-item"
  (with-text-size (s (txtsize *application-frame*))
    (grid s indent item)
    (pointr s)
    (draw-icon s item)
    (call-next-method)  ; <- item-name
    (when (show-children item)
        (dolist (child (children item))
          (disp-tree child pt s (1+ indent))))))

;leaves
(defmethod disp-tree :before ((item leaf) pt s indent)
  "This displays grid and pointer of a leaf-item"
  (grid s indent item)
  (pointr s))

;names
(defmethod disp-tree (item pt s indent)
  "This presents the item's name"
  (stream-increment-cursor-position s (* 1.5 d) nil)
  (present (item-name item) pt :stream s) (terpri s))

;************************************************************
; 4) A GENERIC APPLICATION FOR VIEWING A TREE
;************************************************************
(define-application-frame tree ()
  ((txtsize :accessor txtsize :initform :normal)
   (group :accessor group :initarg :group :documentation "group is an object of a node-class")
   (ptype :accessor ptype :initarg :ptype))
  (:pane (make-pane 'application-pane :display-function 'display-tree :incremental-redisplay t :end-of-line-action :allow :end-of-page-action :allow)))

(defmethod display-tree ((f tree) p) (disp-tree (group f) (ptype f) p 0))
(define-presentation-action toggle (icon command tree) (object window) (toggle object) (redisplay-frame-pane *application-frame* window))
(define-tree-command (txt-size :menu t) () (setf (txtsize *application-frame*) textsize))

(defun tree-view (group &optional (frame 'tree) (ptype 'string) &key (left 0) (top 0) (right 400) (bottom 400)(pretty-name (symbol-name frame)) &allow-other-keys)
  (run-frame-top-level (make-application-frame frame :group group :ptype ptype :left left :top top :right right :bottom bottom :pretty-name pretty-name)))

;************************************************************
; 5) SOME FUNCTIONS FOR SIMPLE CASES
;************************************************************
;-----------------
; 1) *** VIEW A STRING- or SYMBOL-TREE of this form (using a hash-table)
; '(("node1"            ; '((node1 
;    ("node11"          ;    (node11 
;     ("leaf111")       ;     (leaf111)
;     ("leaf112"))))    ;     (leaf112))))

; -1- string-items
(define-node-methods)

;; include autoscroll
; rename 
;;;list-view
;view-list  VIEW LIST <----  view-lol??    items may be strings or symbols 
(defun treeview (tree &optional (key (caar tree)))
  "view a tree"  ; of stringitems  <--??
  (t2h tree)  ; 1) create hash-table
  (tree-view (make-instance 'node :name key :show-children t)))

;----------------------------------------------------------------------------------
; 1) *** VIEW DIRECTORY - from http://osdir.com/ml/mcclim-devel/2009-08/msg00010.html
#|
; with cl-fad
(define-node-methods 
  :nc node-fs
  :cc pathname
  :cy path:=
  :nn (let ((lst (pathname-directory (name n)))) (when (consp lst) (car (last lst))))
  :ln (file-namestring (name n))
  :gc  (fad:list-directory (name n))
  :cp (path:-d n))

(defun list-dir (d) ;initial key
  (tree-view (make-instance 'node-fs 
                            :name (path:dirname d) 
                            :show-children t)))
|#

; include scrolling
; without cl-fad
(define-node-methods 
  :nc node-fs
  :cc pathname
  :cy uiop:pathname-equal
  :nn (let ((lst (pathname-directory (name n)))) (when (consp lst) (car (last lst))))
  :ln (file-namestring (name n))
;  :gc  (fad:list-directory (name n)) 

;  :gc (uiop:directory-files (name n)) 
;  :gc (directory (name n))
;  :gc (directory (directory-namestring (name n)))
  :gc (append (uiop:subdirectories (name n)) (uiop:directory-files (name n)))
  :cp (uiop:directory-pathname-p n))

(defun list-dir (d) ;initial key
  (tree-view (make-instance 'node-fs 
                            :name d   ; :name (directory-namestring d) ; geht auch
                            :show-children t)))

#|
;----- test -----
;----------------------------------------------------------------------------------
;scheint nicht zu gehen
(define-node-methods 
  ;:nc node-fs
  :cc pathname
  :cy path:=
  :nn (let ((lst (pathname-directory (name n)))) (when (consp lst) (car (last lst))))
  :ln (file-namestring (name n))
  :gc  (fad:list-directory (name n))
  :cp (path:-d n))

(defun list-dir-xx (d) ;initial key
  (tree-view (make-instance 'node ; 'node-fs 
                            :name (path:dirname d) 
                            :show-children t)))



;========================================
;---- with uiop
(define-node-methods 
  :nc node-fs
  :cc pathname
;  :cy path:=
  :cy uiop:pathname-equal
  :nn (let ((lst (pathname-directory (name n)))) (when (consp lst) (car (last lst))))
  :ln (file-namestring (name n))
;  :c  (fad:list-directory (name n))
;  :c  (uiop:directory* (name n))
;  :c  (uiop:directory-files (name n))
  :c  (format t "~s*" (uiop:directory* (name n)))
;  :cp (path:-d n))
  :cp (uiop:directory-exists-p n))

(defun list-dir (d) ;initial key
  (tree-view (make-instance 'node-fs 
;                            :name (path:dirname d) 
                            :name d 
                            :show-children t)))
=======================================================
|#


;(cw:treeview-strings cw-examples::stgtree)

; -1- symbol-items --- converts symbols to strings
;  (define-node-methods
;    :nc node-z)
;  
;  (defun treeview-symbols (tree key)
;    (t2h (sym2stg tree))
;    (tree-view (make-instance 'node-z :name (string-downcase (symbol-name key)) :show-children t) 'string))
;  ;  (tree-view (make-instance 'node-z :name (string-downcase (symbol-name key)) :show-children t) 'string 'string))
;  
;  ;run (cw:treeview-symbols cw-examples::symtree 'icd)
;  ;------------------------------------------------------

;3)view-xml  VIEW XML 
#|
* (s-xml:parse-xml-file "books.xml")

(:|bookstore|
 ((:|book| :|category| "cooking") ((:|title| :|lang| "en") "Everyday Italian")
  (:|author| "Giada De Laurentiis") (:|year| "2005") (:|price| "30.00"))
 ((:|book| :|category| "children") ((:|title| :|lang| "en") "Harry Potter")
  (:|author| "J K. Rowling") (:|year| "2005") (:|price| "29.99"))
 ((:|book| :|category| "web") ((:|title| :|lang| "en") "XQuery Kick Start")
  (:|author| "James McGovern") (:|author| "Per Bothner")
  (:|author| "Kurt Cagle") (:|author| "James Linn")
  (:|author| "Vaidyanathan Nagarajan") (:|year| "2003") (:|price| "49.99"))
 ((:|book| :|category| "web" :|cover| "paperback")
  ((:|title| :|lang| "en") "Learning XML") (:|author| "Erik T. Ray")
  (:|year| "2003") (:|price| "39.95")))
|#

;4) view-debian-packages-all
;   view-executables-available
#|
https://askubuntu.com/questions/598384/how-do-i-get-a-list-of-all-available-packages-from-the-repositories-that-are-con
apt-get update
apt-cache dump | grep -oP 'Package: \K.*' | sort -n          # available

https://askubuntu.com/questions/1056448/how-to-get-the-list-of-installed-packages-on-ubuntu-debian-w-o-command-or-w-o-lo
awk -vRS= '/Status: install/ {print $2}' /var/lib/dpkg/status     #installed

https://superuser.com/questions/492501/how-do-i-list-all-of-the-executables-in-a-linux-directory-sub-dirs-path
find . -type f -executable -maxdepth 1     # find executable files
|#

;;  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  
;;  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  END
;;  
;;  #|
;;  (defun tree-view (gp pt &optional (frame 'tree) &key (left 0) (top 0) (right 400) (bottom 400) pretty-name &allow-other-keys)
;;      (run-frame-top-level (make-application-frame frame :group gp :ptype pt :left left :top top :right right :bottom bottom :pretty-name pretty-name)))
;;  
;;  ;;; siehe zuBehalten QQQQ
;;  ;; oben bug pretty name without default !!! ev use gp or pt ?? statt empty string, 27.8.2019
;;  ;; + problem optional + key !!??
;;  ;; pt ~ frame ?? remove one ??
;;  (defun tree-view (gp pt &optional (frame 'tree) &key (left 0) (top 0) (right 400) (bottom 400) (pretty-name "") &allow-other-keys)
;;      (run-frame-top-level (make-application-frame frame :group gp :ptype pt :left left :top top :right right :bottom bottom :pretty-name pretty-name)))
;;  
;;  ;; 11.10.19 only keys
;;  (defun tree-view (gp pt &key (frame 'tree) (left 0) (top 0) (right 400) (bottom 400) (pretty-name "") &allow-other-keys)
;;      (run-frame-top-level (make-application-frame frame :group gp :ptype pt :left left :top top :right right :bottom bottom :pretty-name pretty-name)))
;;  |#
;;  
;;  ;; 14.10.19
;;  ;(defun tree-view (group &optional (frame 'tree) (ptype 'string) &key &allow-other-keys)
;;  ;    (run-frame-top-level (make-application-frame frame :group group :ptype ptype :left 0 :top 0 :right 400 :bottom 400)))
;;  
;;  
;;  ;######################################################################################################
;;  #|
;;  ;######################################################################################################
;;  ; 16.10.19
;;  ; so geht pkg-doc ru-accent ru-reader atc cw:list-dir 
;;  ; so geht right 800, aber nicht pretty name <--
;;  
;;  (defun tree-view (group &optional (frame 'tree) (ptype 'string) &key (left 0) (top 0) (right 400) (bottom 400) &allow-other-keys)
;;    (run-frame-top-level (make-application-frame frame :group group :ptype ptype :left left :top top :right right :bottom bottom)))
;;  ;######################################################################################################
;;  |#
;;  ;######################################################################################################
;;  
;;  
;;  ;test pretty name, scheint zu gehen, test (icd10lquery::icd-clim :i)   etc
;;  
;;  ;(defun tree-view (group &optional (frame 'tree) (ptype 'string) &rest options  &key (left 0) (top 0) (right 400) (bottom 400) &allow-other-keys)
;;  ;  (run-frame-top-level (make-application-frame frame options :group group :ptype ptype :left left :top top :right right :bottom bottom)))
;;  
;;  
;;  #|
;;  ;;; sieht options in make-application-frame  <--- ???
;;  ;;; so geht pretty name und size 800
;;  (defun tree-view (group &optional (frame 'tree) (ptype 'string) &key (left 0) (top 0) (right 400) (bottom 400) (pretty-name frame) &allow-other-keys)
;;  ;  (run-frame-top-level (make-application-frame frame :group group :ptype ptype :left left :top top :right right :bottom bottom &allow-other-keys)))
;;    (run-frame-top-level (make-application-frame frame :group group :ptype ptype :left left :top top :right right :bottom bottom :pretty-name pretty-name)))
;;  |#
;;  
;;  ;(defun list-dir (d) ;initial key
;;  ;  (tree-view (make-instance 'node-fs 
;;  ;                            :name (path:dirname d) 
;;  ;                            :show-children t) 
;;  ;             'string))
;;  ;
;;  ;  (defun treeview-strings (tree key) ;initial key
;;  ;    (t2h tree)  ; 1) create hash-table
;;  ;    (tree-view (make-instance 'node :name key :show-children t) 'string))            ;; so gehts 9.10.2019
;;  ;  ;  (tree-view (make-instance 'node :name key :show-children t) 'string 'string ))
;;  
;;  ;geht
;;  ;(defun treeview-strings (tree key) ;initial key
;;  ;  (t2h tree)  ; 1) create hash-table
;;  ;  (tree-view (make-instance 'node :name key :show-children t)))
;;  
;;  ;(defun treeview-strings (tree &optional (key (caar tree)))
;(defmethod children (n) nil)
;(defmethod node-p (n) nil)   ; test if child is node
;(defmethod children (n) nil)
;;  
;;  
